"""Polynomials."""

from symfem.polynomials.dual import l2_dual
from symfem.polynomials.jacobi import jacobi_polynomial, monic_jacobi_polynomial
from symfem.polynomials.legendre import orthogonal_basis, orthonormal_basis
from symfem.polynomials.lobatto import lobatto_basis, lobatto_dual_basis
from symfem.polynomials.polysets import (
    Hcurl_polynomials,
    Hcurl_quolynomials,
    Hcurl_serendipity,
    Hdiv_polynomials,
    Hdiv_quolynomials,
    Hdiv_serendipity,
    polynomial_set,
    polynomial_set_1d,
    polynomial_set_vector,
    prism_polynomial_set_1d,
    prism_polynomial_set_vector,
    pyramid_polynomial_set_1d,
    pyramid_polynomial_set_vector,
    quolynomial_set_1d,
    quolynomial_set_vector,
    serendipity_indices,
    serendipity_set_1d,
    serendipity_set_vector,
)
from symfem.polynomials.degree_computation import degree



"""Enriched Galerkin elements.

This element's definition appears in https://doi.org/10.1137/080722953
(Sun, Liu, 2009).
"""

import typing

from symfem.elements.lagrange import Lagrange
from symfem.elements.q import Q
from symfem.finite_element import EnrichedElement
from symfem.references import Reference

__all__ = ["EnrichedGalerkin"]


class EnrichedGalerkin(EnrichedElement):
    """An enriched Galerkin element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        if reference.name in ["quadrilateral", "hexahedron"]:
            super().__init__([Q(reference, order), Q(reference, 0)])
        else:
            super().__init__([Lagrange(reference, order), Lagrange(reference, 0)])

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        if self.reference.name in ["quadrilateral", "hexahedron"]:
            return self.order * self.reference.tdim
        return self.order

    names = ["enriched Galerkin", "EG"]
    references = ["interval", "triangle", "quadrilateral", "tetrahedron", "hexahedron"]
    min_order = 1
    continuity = "C0"
    value_type = "scalar"
    last_updated = "2023.05"



"""Arnold-Winther elements on simplices.

Thse elements definitions appear in https://doi.org/10.1007/s002110100348
(Arnold, Winther, 2002) [conforming] and https://doi.org/10.1142/S0218202503002507
(Arnold, Winther, 2003) [nonconforming]
"""

import typing

import sympy

from symfem.elements.lagrange import Lagrange
from symfem.finite_element import CiarletElement
from symfem.functionals import (
    InnerProductIntegralMoment,
    IntegralMoment,
    ListOfFunctionals,
    PointInnerProduct,
)
from symfem.functions import FunctionInput
from symfem.polynomials import polynomial_set_vector
from symfem.references import Reference
from symfem.symbols import x

__all__ = ["ArnoldWinther", "NonConformingArnoldWinther"]


class ArnoldWinther(CiarletElement):
    """An Arnold-Winther element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert reference.name == "triangle"
        self.variant = variant
        poly: typing.List[FunctionInput] = []
        poly += [
            ((p[0], p[1]), (p[1], p[2])) for p in polynomial_set_vector(reference.tdim, 3, order)
        ]
        poly += [
            (
                (
                    (order - k + 2) * (order - k + 3) * x[0] ** k * x[1] ** (order - k + 1),
                    -k * (order - k + 3) * x[0] ** (k - 1) * x[1] ** (order - k + 2),
                ),
                (
                    -k * (order - k + 3) * x[0] ** (k - 1) * x[1] ** (order - k + 2),
                    k * (k - 1) * x[0] ** (k - 2) * x[1] ** (order - k + 3),
                ),
            )
            for k in range(order + 2)
        ]
        poly += [
            ((0, x[0] ** (order + 1)), (x[0] ** (order + 1), -(order + 1) * x[0] ** order * x[1])),
            ((0, 0), (0, x[0] ** (order + 1))),
        ]

        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.vertices):
            for d in [[(1, 0), (1, 0)], [(1, 0), (0, 1)], [(0, 1), (0, 1)]]:
                dofs.append(
                    PointInnerProduct(
                        reference, v, d[0], d[1], entity=(0, v_n), mapping="double_contravariant"
                    )
                )
        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            sub_e = Lagrange(sub_ref.default_reference(), order - 1, variant)
            for dof_n, dof in enumerate(sub_e.dofs):
                p = sub_e.get_basis_function(dof_n).get_function()
                for component in [sub_ref.normal(), sub_ref.tangent()]:
                    dofs.append(
                        InnerProductIntegralMoment(
                            reference,
                            p,
                            component,
                            sub_ref.normal(),
                            dof,
                            entity=(1, e_n),
                            mapping="double_contravariant",
                        )
                    )
        sub_e = Lagrange(reference, order - 2, variant)
        for dof_n, dof in enumerate(sub_e.dofs):
            p = sub_e.get_basis_function(dof_n).get_function()
            for component22 in [((1, 0), (0, 0)), ((0, 1), (0, 0)), ((0, 0), (0, 1))]:
                dofs.append(
                    IntegralMoment(
                        reference,
                        tuple(tuple(p * j for j in i) for i in component22),
                        dof,
                        entity=(2, 0),
                        mapping="double_contravariant",
                    )
                )

        if order >= 3:
            sub_e = Lagrange(reference, order - 3, variant)
            for p, dof in zip(sub_e.get_basis_functions(), sub_e.dofs):
                if sympy.Poly(p.as_sympy(), x[:2]).degree() != order - 3:
                    continue
                f = p * x[0] ** 2 * x[1] ** 2 * (1 - x[0] - x[1]) ** 2
                J = [
                    [f.diff(x[1]).diff(x[1]), -f.diff(x[0]).diff(x[1])],
                    [-f.diff(x[1]).diff(x[0]), f.diff(x[0]).diff(x[0])],
                ]
                dofs.append(
                    IntegralMoment(reference, J, dof, entity=(2, 0), mapping="double_contravariant")
                )

        super().__init__(
            reference,
            order,
            poly,
            dofs,
            reference.tdim,
            reference.tdim**2,
            (reference.tdim, reference.tdim),
        )

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    names = ["Arnold-Winther", "AW", "conforming Arnold-Winther"]
    references = ["triangle"]
    min_order = 2
    continuity = "integral inner H(div)"
    value_type = "symmetric matrix"
    last_updated = "2025.03"


class NonConformingArnoldWinther(CiarletElement):
    """A nonconforming Arnold-Winther element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert reference.name == "triangle"
        self.variant = variant
        poly: typing.List[FunctionInput] = []
        poly += [
            ((p[0], p[1]), (p[1], p[2])) for p in polynomial_set_vector(reference.tdim, 3, order)
        ]

        poly += [
            ((0, x[1] ** 2), (x[1] ** 2, -2 * x[1] ** 2)),
            ((-2 * x[0] ** 2, x[0] ** 2), (x[0] ** 2, 0)),
            ((-2 * x[0] * x[1], x[0] * x[1]), (x[0] * x[1], 0)),
            ((x[0] * (x[0] + x[1]), 0), (0, 0)),
            ((x[0] ** 2, 0), (0, x[0] * x[1])),
            ((x[0] ** 2, 0), (0, -(x[1] ** 2))),
        ]

        dofs: ListOfFunctionals = []
        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            sub_e = Lagrange(sub_ref.default_reference(), 1, variant)
            for dof_n, dof in enumerate(sub_e.dofs):
                p = sub_e.get_basis_function(dof_n).get_function()
                for component in [sub_ref.normal(), sub_ref.tangent()]:
                    dofs.append(
                        InnerProductIntegralMoment(
                            reference,
                            p,
                            component,
                            sub_ref.normal(),
                            dof,
                            entity=(1, e_n),
                            mapping="double_contravariant",
                        )
                    )
        sub_e = Lagrange(reference, 0, variant)
        for dof_n, dof in enumerate(sub_e.dofs):
            p = sub_e.get_basis_function(dof_n).get_function()
            for component22 in [((1, 0), (0, 0)), ((0, 1), (0, 0)), ((0, 0), (0, 1))]:
                dofs.append(
                    IntegralMoment(
                        reference,
                        tuple(tuple(p * j for j in i) for i in component22),
                        dof,
                        entity=(2, 0),
                        mapping="double_contravariant",
                    )
                )

        super().__init__(
            reference,
            order,
            poly,
            dofs,
            reference.tdim,
            reference.tdim**2,
            (reference.tdim, reference.tdim),
        )

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order + 1

    names = ["nonconforming Arnold-Winther", "nonconforming AW"]
    references = ["triangle"]
    min_order = 1
    max_order = 1
    continuity = "integral inner H(div)"
    value_type = "symmetric matrix"
    last_updated = "2025.03"



"""Enriched vector Galerkin elements.

This element's definition appears in https://doi.org/10.1016/j.camwa.2022.06.018
(Yi, Hu, Lee, Adler, 2022)
"""

import typing

from symfem.elements.lagrange import VectorLagrange
from symfem.elements.q import VectorQ
from symfem.finite_element import CiarletElement, EnrichedElement
from symfem.functionals import BaseFunctional, IntegralAgainst
from symfem.functions import FunctionInput, VectorFunction
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import x

__all__ = ["Enrichment", "VectorEnrichedGalerkin"]


class Enrichment(CiarletElement):
    """An LF enriched Galerkin element."""

    def __init__(self, reference: Reference):
        """Create the element.

        Args:
            reference: The reference element
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()
        f = VectorFunction(tuple(x[i] - j for i, j in enumerate(reference.midpoint())))
        poly: typing.List[FunctionInput] = [f]
        size = f.dot(f).integral(reference, x)
        dofs: typing.List[BaseFunctional] = [
            IntegralAgainst(
                reference, tuple(i / size for i in f), (reference.tdim, 0), "contravariant"
            )
        ]

        super().__init__(reference, 1, poly, dofs, reference.tdim, reference.tdim)

    @property
    def lagrange_subdegree(self) -> int:
        return -1

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return 1

    @property
    def polynomial_subdegree(self) -> int:
        return -1

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return 1

    names: typing.List[str] = []
    references = ["triangle", "quadrilateral", "tetrahedron", "hexahedron"]
    min_order = 1
    max_order = 1
    continuity = "C0"
    value_type = "vector"
    last_updated = "2023.05"


class VectorEnrichedGalerkin(EnrichedElement):
    """An LF enriched Galerkin element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        if reference.name in ["quadrilateral", "hexahedron"]:
            super().__init__([VectorQ(reference, order), Enrichment(reference)])
        else:
            super().__init__([VectorLagrange(reference, order), Enrichment(reference)])

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        if self.reference.name in ["quadrilateral", "hexahedron"]:
            return self.order * self.reference.tdim
        return self.order

    names = ["enriched vector Galerkin", "locking-free enriched Galerkin", "LFEG"]
    references = ["triangle", "quadrilateral", "tetrahedron", "hexahedron"]
    min_order = 1
    continuity = "C0"
    value_type = "vector"
    last_updated = "2023.05"



"""Alfeld-Sorokina element on a triangle.

This element's definition appears in https://doi.org/10.1007/s10543-015-0557-x
(Alfeld, Sorokina, 2015)
"""

import typing

import sympy

from symfem.finite_element import CiarletElement
from symfem.functionals import DotPointEvaluation, ListOfFunctionals, PointDivergenceEvaluation
from symfem.functions import FunctionInput, VectorFunction
from symfem.piecewise_functions import PiecewiseFunction
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import x

__all__ = ["AlfeldSorokina"]


class AlfeldSorokina(CiarletElement):
    """Alfeld-Sorokina finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 2
        assert reference.name == "triangle"
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(DotPointEvaluation(reference, vs, (1, 0), entity=(0, v_n)))
            dofs.append(DotPointEvaluation(reference, vs, (0, 1), entity=(0, v_n)))
            dofs.append(PointDivergenceEvaluation(reference, vs, entity=(0, v_n)))

        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            dofs.append(DotPointEvaluation(reference, sub_ref.midpoint(), (1, 0), entity=(1, e_n)))
            dofs.append(DotPointEvaluation(reference, sub_ref.midpoint(), (0, 1), entity=(1, e_n)))

        mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*reference.vertices))

        subs = [
            (reference.vertices[0], reference.vertices[1], mid),
            (reference.vertices[1], reference.vertices[2], mid),
            (reference.vertices[2], reference.vertices[0], mid),
        ]

        piece_list = [
            tuple(VectorFunction((p, 0)) for _ in range(3))
            for p in [1, x[0], x[1], x[0] ** 2, x[0] * x[1], x[1] ** 2]
        ]
        piece_list += [
            tuple(VectorFunction((0, p)) for _ in range(3))
            for p in [1, x[0], x[1], x[0] ** 2, x[0] * x[1], x[1] ** 2]
        ]

        piece_list.append(
            (
                VectorFunction((4 * x[0] + 12 * x[1] ** 2 - 4 * x[1], 6 * x[0] * x[1] - 4 * x[1])),
                VectorFunction(
                    (
                        3 * x[0] ** 2 + 2 * x[0] + 4 * x[1] - 1,
                        -3 * x[0] ** 2 + 4 * x[0] - 2 * x[1] - 1,
                    )
                ),
                VectorFunction(
                    (
                        3 * x[0] ** 2 + 6 * x[0] * x[1] + 3 * x[1] ** 2,
                        9 * x[0] ** 2 - 4 * x[0] - 3 * x[1] ** 2,
                    )
                ),
            )
        )
        piece_list.append(
            (
                VectorFunction(
                    (
                        144 * x[0] * x[1] + 10 * x[0] + 12 * x[1] ** 2 - 10 * x[1],
                        -30 * x[0] * x[1] + 36 * x[1] ** 2 - 10 * x[1],
                    )
                ),
                VectorFunction(
                    (
                        -69 * x[0] ** 2 + 104 * x[0] + 46 * x[1] - 25,
                        24 * x[0] ** 2 - 26 * x[0] + 4 * x[1] + 2,
                    )
                ),
                VectorFunction(
                    (39 * x[0] ** 2 + 96 * x[0] * x[1] + 21 * x[1] ** 2, -10 * x[0] + 6 * x[1] ** 2)
                ),
            )
        )
        piece_list.append(
            (
                VectorFunction(
                    (
                        -14 * x[0] + 12 * x[1] ** 2 + 14 * x[1],
                        42 * x[0] * x[1] - 108 * x[1] ** 2 + 14 * x[1],
                    )
                ),
                VectorFunction(
                    (
                        3 * x[0] ** 2 - 16 * x[0] + 22 * x[1] - 1,
                        24 * x[0] ** 2 + 144 * x[0] * x[1] - 50 * x[0] - 92 * x[1] + 26,
                    )
                ),
                VectorFunction(
                    (
                        -33 * x[0] ** 2 + 24 * x[0] * x[1] + 21 * x[1] ** 2,
                        14 * x[0] - 66 * x[1] ** 2,
                    )
                ),
            )
        )

        poly: typing.List[FunctionInput] = []
        poly += [PiecewiseFunction({i: j for i, j in zip(subs, p)}, 2) for p in piece_list]

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    @property
    def lagrange_subdegree(self) -> int:
        raise NotImplementedError()

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        raise NotImplementedError()

    @property
    def polynomial_subdegree(self) -> int:
        raise NotImplementedError()

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        raise NotImplementedError()

    names = ["Alfeld-Sorokina", "AS"]
    references = ["triangle"]
    min_order = 2
    max_order = 2
    continuity = "C0"
    value_type = "scalar macro"
    last_updated = "2023.05"



"""Orthogonal (Legendre) polynomials."""

import typing

import sympy

from symfem.functions import ScalarFunction
from symfem.symbols import AxisVariablesNotSingle, x
from symfem.polynomials.jacobi import jacobi_polynomial

__all__: typing.List[str] = []


def orthogonal_basis_interval(
    order: int, variables: AxisVariablesNotSingle = [x[0]]
) -> typing.List[ScalarFunction]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 1

    return [jacobi_polynomial(i, 0, 0, 2 * variables[0] - 1) for i in range(order + 1)]


def orthogonal_basis_triangle(
    order: int, variables: AxisVariablesNotSingle = [x[0], x[1]]
) -> typing.List[ScalarFunction]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 2

    return [
        jacobi_polynomial(p, 0, 0, 2 * variables[0] / (1 - variables[1]) - 1)
        * (1 - variables[1]) ** p
        * jacobi_polynomial(q, 2 * p + 1, 0, 2 * variables[1] - 1)
        for p in range(order + 1)
        for q in range(order - p + 1)
    ]


def orthogonal_basis_quadrilateral(
    order: int, variables: AxisVariablesNotSingle = [x[0], x[1]]
) -> typing.List[ScalarFunction]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 2
    return [
        a * b
        for a in orthogonal_basis_interval(order, [variables[0]])
        for b in orthogonal_basis_interval(order, [variables[1]])
    ]


def orthogonal_basis_tetrahedron(
    order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 3

    return [
        jacobi_polynomial(p, 0, 0, 2 * variables[0] / (1 - variables[1] - variables[2]) - 1)
        * (1 - variables[1] - variables[2]) ** p
        * jacobi_polynomial(q, 2 * p + 1, 0, 2 * variables[1] / (1 - variables[2]) - 1)
        * (1 - variables[2]) ** q
        * jacobi_polynomial(r, 2 * (p + q + 1), 0, 2 * variables[2] - 1)
        for p in range(order + 1)
        for q in range(order - p + 1)
        for r in range(order - p - q + 1)
    ]


def orthogonal_basis_hexahedron(
    order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 3
    return [
        a * b * c
        for a in orthogonal_basis_interval(order, [variables[0]])
        for b in orthogonal_basis_interval(order, [variables[1]])
        for c in orthogonal_basis_interval(order, [variables[2]])
    ]


def orthogonal_basis_prism(
    order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """Create a basis of orthogonal polynomials.

    Args:
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 3

    return [
        a * b
        for a in orthogonal_basis_triangle(order, [variables[0], variables[1]])
        for b in orthogonal_basis_interval(order, [variables[2]])
    ]


def orthogonal_basis_lagrange_pyramid(
    order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """Create a basis of orthogonal rationomials.

    Args:
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 3

    return [
        jacobi_polynomial(p, 0, 0, (2 * variables[0] + variables[2] - 1) / (1 - variables[2]))
        * jacobi_polynomial(q, 0, 0, (2 * variables[1] + variables[2] - 1) / (1 - variables[2]))
        * (1 - variables[2]) ** max(p, q)
        * jacobi_polynomial(r, 2 * max(p, q) + 2, 0, 2 * variables[2] - 1)
        for r in range(order + 1)
        for p in range(order + 1 - r)
        for q in range(order + 1 - r)
    ]


def orthogonal_basis_full_pyramid(
    order: int, variables: AxisVariablesNotSingle = x
) -> typing.List[ScalarFunction]:
    """Create a basis of orthogonal rationomials.

    Args:
        order: The maximum polynomial degree
        variables: The variables to use

    Returns:
        A set of orthogonal polynomials
    """
    assert len(variables) == 3

    return [
        jacobi_polynomial(p, 0, 0, (2 * variables[0] + variables[2] - 1) / (1 - variables[2]))
        * jacobi_polynomial(q, 0, 0, (2 * variables[1] + variables[2] - 1) / (1 - variables[2]))
        * jacobi_polynomial(r, 2, 0, 2 * variables[2] - 1)
        for p in range(order + 1)
        for q in range(order + 1)
        for r in range(order + 1)
    ]


def orthogonal_basis(
    cell: str,
    order: int,
    variables: typing.Optional[AxisVariablesNotSingle] = None,
    ptype: str = "Lagrange",
) -> typing.List[ScalarFunction]:
    """Create a basis of orthogonal polynomials.

    Args:
        cell: The cell type
        order: The maximum polynomial degree
        variables: The variables to use
        ptype: The type of the polynomial set (pyramids only)

    Returns:
        A set of orthogonal polynomials
    """
    args: typing.List[typing.Any] = [order]
    if variables is not None:
        args.append(variables)

    if ptype != "Lagrange" and cell != "pyramid":
        raise ValueError("Cannot use ptype input for non-pyramid cells")

    if cell == "interval":
        return orthogonal_basis_interval(*args)
    if cell == "triangle":
        return orthogonal_basis_triangle(*args)
    if cell == "quadrilateral":
        return orthogonal_basis_quadrilateral(*args)
    if cell == "hexahedron":
        return orthogonal_basis_hexahedron(*args)
    if cell == "tetrahedron":
        return orthogonal_basis_tetrahedron(*args)
    if cell == "prism":
        return orthogonal_basis_prism(*args)
    if cell == "pyramid":
        if ptype == "Lagrange":
            return orthogonal_basis_lagrange_pyramid(*args)
        if ptype == "full":
            return orthogonal_basis_full_pyramid(*args)

    raise ValueError(f"Unsupported cell type: {cell}")


def orthonormal_basis(
    cell: str,
    order: int,
    variables: typing.Optional[AxisVariablesNotSingle] = None,
    ptype: str = "Lagrange",
) -> typing.List[ScalarFunction]:
    """Create a basis of orthonormal polynomials.

    Args:
        cell: The cell type
        order: The maximum polynomial degree
        variables: The variables to use
        ptype: The type of the polynomial set (pyramids only)

    Returns:
        A set of orthonormal polynomials
    """
    from symfem.create import create_reference

    if ptype != "Lagrange" and cell != "pyramid":
        raise ValueError("Cannot use ptype input for non-pyramid cells")

    poly = orthogonal_basis(cell, order, variables, ptype)
    ref = create_reference(cell)
    if variables is None:
        variables = x
    norms = [sympy.sqrt((f**2).integral(ref, dummy_vars=variables)) for f in poly]
    for i, n in enumerate(norms):
        poly[i] /= n
    return poly



"""Interface to Basix."""

import typing
from enum import Enum
from symfem.finite_element import FiniteElement, CiarletElement
from symfem.symbols import x
from symfem.piecewise_functions import PiecewiseFunction
from symfem.polynomials import degree

import numpy as np
import numpy.typing as npt
import basix
import basix.ufl

sobolev_spaces = {
    "L2": basix.SobolevSpace.L2,
    "C0": basix.SobolevSpace.H1,
    "C1": basix.SobolevSpace.H2,
    "C2": basix.SobolevSpace.H3,
    "H(div)": basix.SobolevSpace.HDiv,
    "H(curl)": basix.SobolevSpace.HCurl,
    "inner H(div)": basix.SobolevSpace.HDivDiv,
    "integral inner H(div)": basix.SobolevSpace.HDivDiv,
    "inner H(curl)": basix.SobolevSpace.HEin,
}

map_types = {
    "identity": basix.MapType.identity,
    "covariant": basix.MapType.covariantPiola,
    "contravariant": basix.MapType.contravariantPiola,
    "double_covariant": basix.MapType.doubleCovariantPiola,
    "double_contravariant": basix.MapType.doubleContravariantPiola,
}


def get_embedded_degrees(poly, reference) -> typing.Tuple[int, int]:
    """Get embedded degrees of a set of polynomials.

    Args:
        poly: Polynomials
        reference: Reference cell

    Returns: Embedded sub- and superdegrees
    """
    superdegree = max(degree(reference, p) for p in poly)
    return (-1, superdegree)


def _create_custom_element_args(
    element: FiniteElement, dtype: npt.DTypeLike = np.float64
) -> tuple[list[typing.Any], dict[str, typing.Any]]:
    """Generate the arguments to create a Basix custom element.

    Args:
        element: The Symfem element
        dtype: The dtype of the Basix element

    Returns:
        A list of args and a dict of kwargs
    """
    if not isinstance(element, CiarletElement):
        raise NotImplementedError("Can only convert Ciarlet elements to Basix custom elements")
    for dof in element.dofs:
        if dof.nderivs > 0:
            raise NotImplementedError(
                "Conversion to Basix element not implemented for elements including derivatives"
            )
    for dof in element.dofs[1:]:
        if dof.mapping != element.dofs[0].mapping:
            raise NotImplementedError(
                "Conversion to Basix element not implemented for elements with a mixture of mapping types"
            )
    for p in element.get_polynomial_basis():
        if isinstance(p, PiecewiseFunction):
            raise NotImplementedError(
                "Conversion to Basix element not implemented for macro elements"
            )

    map_name = element.dofs[0].mapping
    continuity = element.continuity
    assert map_name is not None
    assert continuity is not None
    map_type = map_types[map_name]
    sobolev_space = sobolev_spaces[continuity.replace("{order}", f"{element.order}")]

    poly = element.get_polynomial_basis()
    subdegree, superdegree = get_embedded_degrees(poly, element.reference)
    cell = getattr(basix.CellType, element.reference.name)
    ptype = basix.PolysetType.standard
    nderivs = max(dof.nderivs for dof in element.dofs)

    pts, wts = basix.make_quadrature(cell, superdegree * 2)
    opoly = basix.polynomials.tabulate_polynomial_set(cell, ptype, superdegree, 0, pts)[0]

    wcoeffs = np.empty((len(poly), opoly.shape[0] * element.range_dim))
    if element.range_dim == 1:
        for i, p in enumerate(poly):
            values = np.array([float(p.subs(x, list(pt))) for pt in pts])
            for j, q in enumerate(opoly):
                wcoeffs[i, j] = (q * values * wts).sum()
    else:
        for i, p in enumerate(poly):
            if p.is_vector:
                for c in range(element.range_dim):
                    values = np.array([float(p[c].subs(x, list(pt))) for pt in pts])
                    for j, q in enumerate(opoly):
                        wcoeffs[i, c * len(opoly) + j] = (q * values * wts).sum()
            elif p.is_matrix:
                rshape = element.range_shape
                assert rshape is not None
                assert len(rshape) == 2
                for c0 in range(rshape[0]):
                    for c1 in range(rshape[1]):
                        values = np.array([float(p[c0][c1].subs(x, list(pt))) for pt in pts])
                        for j, q in enumerate(opoly):
                            wcoeffs[i, (c0 * rshape[1] + c1) * len(opoly) + j] = (
                                q * values * wts
                            ).sum()
            else:
                raise NotImplementedError(f"Unsupported polynomial type: {type(p)}")

    ref = element.reference
    dof_pts = [
        [np.empty((0, ref.gdim), dtype=dtype) for _ in ref.sub_entities(dim)] for dim in range(4)
    ]
    dof_wts = [
        [
            np.empty((0, element.range_dim, 0, ref.derivative_count(nderivs)), dtype=dtype)
            for _ in ref.sub_entities(dim)
        ]
        for dim in range(4)
    ]

    for dof in element.dofs:
        dof_p, _dof_w = dof.discrete(superdegree)
        dof_w = np.array(_dof_w)
        shape = dof_wts[dof.entity[0]][dof.entity[1]].shape
        new_dof_wts = np.zeros((shape[0] + 1, shape[1], shape[2], shape[3]))
        new_dof_wts[: shape[0], :, :, :] = dof_wts[dof.entity[0]][dof.entity[1]]
        for p_i, pt in enumerate(dof_p):
            dof_wts[dof.entity[0]][dof.entity[1]] = new_dof_wts
            for i, q in enumerate(dof_pts[dof.entity[0]][dof.entity[1]]):
                if np.allclose(pt, q):
                    point_n = i
                    break
            else:
                point_n = dof_pts[dof.entity[0]][dof.entity[1]].shape[0]

                new_dof_pts = np.zeros((point_n + 1, ref.gdim))
                new_dof_pts[:point_n, :] = dof_pts[dof.entity[0]][dof.entity[1]]
                new_dof_pts[point_n, :] = pt
                dof_pts[dof.entity[0]][dof.entity[1]] = new_dof_pts

                shape = dof_wts[dof.entity[0]][dof.entity[1]].shape
                new_dof_wts = np.zeros((shape[0], shape[1], shape[2] + 1, shape[3]))
                new_dof_wts[:, :, : shape[2], :] = dof_wts[dof.entity[0]][dof.entity[1]]
                dof_wts[dof.entity[0]][dof.entity[1]] = new_dof_wts
            dof_wts[dof.entity[0]][dof.entity[1]][-1, :, point_n, :] = dof_w[:, p_i, :]

    return [
        cell,
        () if element.range_shape is None else element.range_shape,
        wcoeffs,
        dof_pts,
        dof_wts,
        nderivs,
        map_type,
        sobolev_space,
        False,
        subdegree,
        superdegree,
        ptype,
    ], {"dtype": dtype}


def create_basix_element(
    element: FiniteElement, dtype: npt.DTypeLike = np.float64, ufl: bool = False
) -> typing.Union[basix.finite_element.FiniteElement, basix.ufl._ElementBase]:
    """Create a Basix element from a Symfem element.

    Args:
        element: The Symfem element
        dtype: The dtype of the Basix element
        ufl: If generating Python, a basix.ufl element will be created if this is set to True

    Returns:
        A Basix element
    """
    args, kwargs = _create_custom_element_args(element, dtype)
    if ufl:
        return basix.ufl.custom_element(*args, **kwargs)
    else:
        return basix.create_custom_element(*args, **kwargs)


def _to_python_string(item: typing.Any, in_array: bool = False) -> str:
    if isinstance(item, np.ndarray):
        if item.size == 0:
            return f"np.empty({_to_python_string(item.shape)}, dtype=np.{item.dtype})"
        if in_array:
            return "[" + ", ".join(_to_python_string(i, True) for i in item) + "]"
        else:
            return (
                "np.array(["
                + ", ".join(_to_python_string(i, True) for i in item)
                + f"], dtype=np.{item.dtype})"
            )
    if isinstance(item, tuple):
        if len(item) == 1:
            return f"({_to_python_string(item[0])}, )"
        return "(" + ", ".join(_to_python_string(i) for i in item) + ")"
    if isinstance(item, list):
        return "[" + ", ".join(_to_python_string(i) for i in item) + "]"
    if isinstance(item, set):
        return "{" + ", ".join(_to_python_string(i) for i in item) + "}"
    if isinstance(item, dict):
        return (
            "{"
            + ", ".join(f"{_to_python_string(i)}: {_to_python_string(j)}" for i, j in item.items())
            + "}"
        )
    if isinstance(item, basix.CellType):
        return f"basix.CellType.{item.name}"
    if isinstance(item, basix.ElementFamily):
        return f"basix.ElementFamily.{item.name}"
    if isinstance(item, basix.MapType):
        return f"basix.MapType.{item.name}"
    if isinstance(item, basix.SobolevSpace):
        return f"basix.SobolevSpace.{item.name}"
    if isinstance(item, basix.PolysetType):
        return f"basix.PolysetType.{item.name}"
    if isinstance(item, Enum):
        raise NotImplementedError(f"Invalid item type: {type(item)}")
    if isinstance(item, (float, bool, int)):
        return f"{item}"
    if isinstance(item, type):
        if item.__module__ == "numpy":
            return f"np.{item.__name__}"
        return f"{item.__module__}.{item.__name__}"

    raise NotImplementedError(f"Invalid item type: {type(item)}")


def _to_cpp_string(item: typing.Any, in_array: bool = False) -> tuple[list[str], str]:
    if isinstance(item, (tuple, list)):
        return [], "{" + ", ".join(_to_cpp_string(i)[1] for i in item) + "}"
    if isinstance(item, (set, dict, np.ndarray)):
        raise NotImplementedError(f"Invalid item type: {type(item)}")
    if isinstance(item, basix.CellType):
        return [], f"basix::cell::type::{item.name}"
    if isinstance(item, basix.ElementFamily):
        return [], f"basix::element::family::{item.name}"
    if isinstance(item, basix.MapType):
        return [], f"basix::maps::type::{item.name}"
    if isinstance(item, basix.SobolevSpace):
        return [], f"basix::sobolev::space::{item.name}"
    if isinstance(item, basix.PolysetType):
        return [], f"basix::polyset::type::{item.name}"
    if isinstance(item, Enum):
        raise NotImplementedError(f"Invalid item type: {type(item)}")
    if isinstance(item, bool):
        return [], "true" if item else "false"
    if isinstance(item, (float, bool, int)):
        return [], f"{item}"

    raise NotImplementedError(f"Invalid item type: {type(item)}")


def generate_basix_element_code(
    element: FiniteElement,
    language: str = "python",
    dtype: typing.Union[npt.DTypeLike, str] = np.float64,
    variable_name: str = "e",
    *,
    ufl: typing.Optional[bool] = None,
    include_imports: bool = True,
    include_comment: bool = True,
) -> str:
    """Generate code to create a Basix custom element.

    Args:
        element: The Symfem element
        language: Programming language to generate ("python" or "c++")
        dtype: The dtype of the Basix element
        variable_name: The variable name to use for the element in the code
        ufl: If generating Python, a basix.ufl element will be created if this is set to True
        include_imports: Should imports be included in the generated code?
        include_comment: Should comment saying what is being created be included in the generated code?

    Returns:
        A Basix element
    """
    args, kwargs = _create_custom_element_args(element, dtype)
    if language == "python":
        code = ""
        if include_imports:
            code += "import basix\n"
            if ufl:
                code += "import basix.ufl\n"
            code += "import numpy as np\n"
            code += "\n"
        if include_comment:
            code += f"# Create degree {element.lagrange_superdegree} {element.name} element\n"
        if ufl:
            code += f"{variable_name} = basix.ufl.custom_element(\n    "
        else:
            code += f"{variable_name} = basix.create_custom_element(\n    "
        code += (
            ",\n    ".join(_to_python_string(i) for i in args)
            + "".join(f", {j}={_to_python_string(k)}" for j, k in kwargs.items())
            + "\n)\n"
        )
        return code

    if language == "c++":
        if ufl is not None:
            raise ValueError("ufl option cannot be used when generating C++")

        if isinstance(dtype, str):
            t = dtype
        elif dtype == np.float64:
            t = "double"
        elif dtype == np.float32:
            t = "float"
        else:
            raise ValueError(f"Invalid dtype: {dtype}")

        definitions = []

        function_args = []

        for n, a in enumerate(args):
            if n == 2:
                f = "wcoeffs"
                data = f"std::vector<{t}> data = {{"
                data += ", ".join(_to_cpp_string(c)[1] for b in a for c in b)
                data += "};"
                d = [
                    data,
                    f"basix::impl::mdspan_t<const {t}, 2> wcoeffs(data.data(), {a.shape[0]}, {a.shape[1]});",
                ]
            elif n == 3:
                d = [f"std::array<std::vector<basix::impl::mdarray_t<{t}, 2>>, 4> x;"]
                for i, x_i in enumerate(a):
                    for x_ij in x_i:
                        if x_ij.size == 0:
                            d.append(f"x[{i}].emplace_back({x_ij.shape[0]}, {x_ij.shape[1]});")
                        else:
                            d.append("{")
                            d.append(
                                f"  auto &_x = x[{i}].emplace_back({x_ij.shape[0]}, {x_ij.shape[1]});"
                            )
                            for k, x_ijk in enumerate(x_ij):
                                for m, x_ijkm in enumerate(x_ijk):
                                    d.append(f"  _x({k}, {m}) = {_to_cpp_string(x_ijkm)[1]};")
                            d.append("}")
                d.append(
                    f"std::array<std::vector<basix::impl::mdspan_t<const {t}, 2>>, 4> xview = basix::impl::to_mdspan(x);"
                )
                f = "xview"
            elif n == 4:
                d = [f"std::array<std::vector<basix::impl::mdarray_t<{t}, 4>>, 4> M;"]
                for i, m_i in enumerate(a):
                    for m_ij in m_i:
                        if m_ij.size == 0:
                            d.append(
                                f"M[{i}].emplace_back({m_ij.shape[0]}, {m_ij.shape[1]}, {m_ij.shape[2]}, {m_ij.shape[3]});"
                            )
                        else:
                            d.append("{")
                            d.append(
                                f"  auto &_M = M[{i}].emplace_back({m_ij.shape[0]}, {m_ij.shape[1]}, {m_ij.shape[2]}, {m_ij.shape[3]});"
                            )
                            for k, m_ijk in enumerate(m_ij):
                                for m, m_ijkm in enumerate(m_ijk):
                                    for n, m_ijkmn in enumerate(m_ijkm):
                                        for o, m_ijkmno in enumerate(m_ijkmn):
                                            d.append(
                                                f"  _M({k}, {m}, {n}, {o}) = {_to_cpp_string(m_ijkmno)[1]};"
                                            )
                            d.append("}")
                d.append(
                    f"std::array<std::vector<basix::impl::mdspan_t<const {t}, 4>>, 4> Mview = basix::impl::to_mdspan(M);"
                )
                f = "Mview"
            else:
                d, f = _to_cpp_string(a)
            definitions += d
            function_args.append(f)

        code = ""
        if include_imports:
            code += "#include <basix/finite-element.h>\n"
            code += "#include <vector>\n"
            code += "\n"
        code += "\n".join(definitions) + "\n"
        code += "\n"
        if include_comment:
            code += f"// Create degree {element.lagrange_superdegree} {element.name} element\n"
        code += f"auto {variable_name} = basix::create_custom_element(\n"
        code += ",\n".join(f"  {a}" for a in function_args) + "\n"
        code += ");\n"

        return code

    raise NotImplementedError(f"Unsupported language: {language}")



"""Functionals used to define the dual sets."""

import typing
from abc import ABC, abstractmethod, abstractproperty

import sympy

from symfem import mappings
from symfem.functions import (
    Function,
    FunctionInput,
    MatrixFunction,
    ScalarFunction,
    VectorFunction,
    parse_function_input,
)
from symfem.geometry import PointType, SetOfPoints
from symfem.polynomials import degree
from symfem.piecewise_functions import PiecewiseFunction
from symfem.quadrature import numerical as numerical_quadrature
from symfem.references import Interval, NonDefaultReferenceError, Reference
from symfem.symbols import t, x

ScalarValueOrFloat = typing.Union[sympy.core.expr.Expr, float]
DiscreteDof = typing.Tuple[
    typing.List[typing.List[float]], typing.List[typing.List[typing.List[float]]]
]

__all__ = [
    "BaseFunctional",
    "PointEvaluation",
    "WeightedPointEvaluation",
    "DerivativePointEvaluation",
    "PointDirectionalDerivativeEvaluation",
    "PointNormalDerivativeEvaluation",
    "PointComponentSecondDerivativeEvaluation",
    "PointInnerProduct",
    "DotPointEvaluation",
    "PointDivergenceEvaluation",
    "IntegralAgainst",
    "IntegralOfDivergenceAgainst",
    "IntegralOfDirectionalMultiderivative",
    "IntegralMoment",
    "DerivativeIntegralMoment",
    "DivergenceIntegralMoment",
    "TangentIntegralMoment",
    "NormalIntegralMoment",
    "TraceIntegralMoment",
    "NormalDerivativeIntegralMoment",
    "InnerProductIntegralMoment",
    "NormalInnerProductIntegralMoment",
]


def _to_tex(f: FunctionInput, tfrac: bool = False) -> str:
    r"""Convert an expresson to TeX.

    Args:
        f: the function
        tfrac: Should \\tfrac be used in the place of \\frac?

    Returns:
        The function as TeX
    """
    out = parse_function_input(f).as_tex()

    if tfrac:
        return out.replace("\\frac", "\\tfrac")
    else:
        return out


def _nth(n: int) -> str:
    """Add th, st or nd to a number.

    Args:
        n: The number

    Returns:
        The string (n)th, (n)st, (n)rd or (n)nd
    """
    if n % 10 == 1 and n != 11:
        return f"{n}st"
    if n % 10 == 2 and n != 12:
        return f"{n}nd"
    if n % 10 == 3 and n != 12:
        return f"{n}rd"
    return f"{n}th"


def discrete_integral_moment(
    domain: Reference, f: Function, poly_degree: int, vars: typing.Tuple[sympy.Symbol, ...] = t
):
    """Get points and weights that define an integral moment against f.

    Args:
        domain: The integral domain
        f: The function to integrate against
        poly_degree: The polynomial degree of the element. This may be used to decide which
            degree quadrature rule to use

    Returns:
        Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
    """
    pts, wts = numerical_quadrature(domain.name, poly_degree + degree(domain, f, vars))
    mapped_pts = [
        [
            float(i[0] + sum(j * k for j, k in zip(i[1:], p)))
            for i in zip(domain.origin, *domain.axes)
        ]
        for p in pts
    ]

    dof_wts = []

    if f.is_scalar:
        sub_w = []
        for p, w in zip(pts, wts):
            value = f.subs(t[: len(p)], p)
            sub_w.append([float(value) * w])
        dof_wts.append(sub_w)
    elif f.is_vector:
        for i in range(f.shape[0]):
            sub_w = []
            for p, w in zip(pts, wts):
                value = f.subs(t[: len(p)], p)
                sub_w.append([float(value[i]) * w])
            dof_wts.append(sub_w)
    else:
        assert f.is_matrix
        for i0 in range(f.shape[0]):
            for i1 in range(f.shape[1]):
                sub_w = []
                for p, w in zip(pts, wts):
                    value = f.subs(t[: len(p)], p)
                    sub_w.append([float(value[i0][i1]) * w])
                dof_wts.append(sub_w)

    return mapped_pts, dof_wts


class BaseFunctional(ABC):
    """A functional."""

    def __init__(
        self, reference: Reference, entity: typing.Tuple[int, int], mapping: typing.Union[str, None]
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        self.reference = reference
        self.entity = entity
        self.mapping = mapping

    @abstractproperty
    def nderivs(self) -> int:
        """Number of derivatives."""

    def entity_dim(self) -> int:
        """Get the dimension of the entity this DOF is associated with.

        Returns:
            The dimension of the entity this DOF is associated with
        """
        return self.entity[0]

    def entity_number(self) -> int:
        """Get the number of the entity this DOF is associated with.

        Returns:
            The number of the entity this DOF is associated with
        """
        return self.entity[1]

    def perform_mapping(
        self, fs: typing.List[Function], map: PointType, inverse_map: PointType
    ) -> typing.List[Function]:
        """Map functions to a cell.

        Args:
            fs: functions
            map: Map from the reference cell to a physical cell
            inverse_map: Map to the reference cell from a physical cell

        Returns:
            Mapped functions
        """
        assert self.mapping is not None
        return [getattr(mappings, self.mapping)(f, map, inverse_map) for f in fs]

    def entity_tex(self) -> str:
        """Get the entity the DOF is associated with in TeX format.

        Returns:
            TeX representation of entity
        """
        if self.entity[0] == self.reference.tdim:
            return "R"
        else:
            return f"{'vefc'[self.entity[0]]}_{{{self.entity[1]}}}"

    def entity_definition(self) -> str:
        """Get the definition of the entity the DOF is associated with.

        Returns:
            The definition
        """
        if self.entity[0] == self.reference.tdim:
            return "\\(R\\) is the reference element"
        else:
            desc = f"\\({self.entity_tex()}\\) is the {_nth(self.entity[1])} "
            desc += ["vertex", "edge", "face", "volume"][self.entity[0]]
            return desc

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        return None

    def eval(
        self, function: Function, symbolic: bool = True
    ) -> typing.Union[ScalarFunction, float]:
        """Apply to the functional to a function.

        Args:
            function: The function
            symbolic: Should it be applied symbolically?

        Returns:
            The value of the functional for the function
        """
        value = self.eval_symbolic(function)
        if symbolic:
            return value
        else:
            return float(value)

    @abstractmethod
    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pass

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        point = self.dof_point()
        if self.entity[0] == 0:
            return point

        midpoint = self.reference.sub_entity(*self.entity).midpoint()
        return tuple(m + sympy.Rational(7, 10) * (p - m) for m, p in zip(midpoint, point))

    def eval_symbolic(self, function: Function) -> ScalarFunction:
        """Symbolically apply the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        e = self._eval_symbolic(function)
        assert isinstance(e, ScalarFunction)
        es = e.as_sympy()
        assert isinstance(es, sympy.core.expr.Expr)
        assert es.is_constant()
        return e

    @abstractmethod
    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        pass

    @abstractmethod
    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        pass

    @abstractmethod
    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """

    name = "Base functional"


class PointEvaluation(BaseFunctional):
    """A point evaluation."""

    def __init__(
        self,
        reference: Reference,
        point_in: FunctionInput,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 0

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        return function.subs(x, self.point)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.point, VectorFunction)
        return f"v\\mapsto v({','.join([_to_tex(i, True) for i in self.point])})", []

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        return [[float(i) for i in self.dof_point()]], [[[1.0]]]

    name = "Point evaluation"


class WeightedPointEvaluation(BaseFunctional):
    """A point evaluation."""

    def __init__(
        self,
        reference: Reference,
        point_in: FunctionInput,
        weight: sympy.core.expr.Expr,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            weight: The weight
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.weight = weight

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 0

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        return function.subs(x, self.point) * self.weight

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.point, VectorFunction)
        return (
            f"v\\mapsto {_to_tex(self.weight)} "
            f"v({','.join([_to_tex(i, True) for i in self.point])})"
        ), []

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        return [[float(i) for i in self.dof_point()]], [[[float(self.weight)]]]

    name = "Weighted point evaluation"


class DerivativePointEvaluation(BaseFunctional):
    """A point evaluation of a given derivative."""

    def __init__(
        self,
        reference: Reference,
        point_in: FunctionInput,
        derivative: typing.Tuple[int, ...],
        entity: typing.Tuple[int, int],
        mapping: typing.Optional[str] = None,
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            derivative: The order(s) of the derivative
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.derivative = derivative

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return sum(self.derivative)

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        for i, j in zip(x, self.derivative):
            for k in range(j):
                function = function.diff(i)
        return function.subs(x, self.point)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def perform_mapping(
        self, fs: typing.List[Function], map: PointType, inverse_map: PointType
    ) -> typing.List[Function]:
        """Map functions to a cell.

        Args:
            fs: functions
            map: Map from the reference cell to a physical cell
            inverse_map: Map to the reference cell from a physical cell

        Returns:
            Mapped functions
        """
        raise mappings.MappingNotImplemented()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.point, VectorFunction)
        if len(self.point) == 1 and self.derivative[0] == 1:
            desc = "v\\mapsto "
            desc += f"v'({','.join([_to_tex(i, True) for i in self.point])})"
            return desc, []
        desc = "v\\mapsto"
        desc += "\\frac{\\partial"
        if sum(self.derivative) > 1:
            desc += f"^{{{sum(self.derivative)}}}"
        desc += "}{"
        for v, i in zip("xyz", self.derivative):
            if i > 0:
                desc += f"\\partial {v}"
                if i > 1:
                    desc += f"^{{{i}}}"
        desc += "}"
        desc += f"v({','.join([_to_tex(i, True) for i in self.point])})"
        return desc, []

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Point derivative evaluation"


class PointDirectionalDerivativeEvaluation(BaseFunctional):
    """A point evaluation of a derivative in a fixed direction."""

    def __init__(
        self,
        reference: Reference,
        point_in: FunctionInput,
        direction_in: FunctionInput,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The points
            direction_in: The diection
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.dir = parse_function_input(direction_in)

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 1

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        d = self.dir.as_sympy()
        assert isinstance(d, tuple)
        return function.directional_derivative(d).subs(x, self.dof_point())

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        d = self.dir.as_sympy()
        assert isinstance(d, tuple)
        return d

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.point, VectorFunction)
        assert isinstance(self.dir, VectorFunction)
        if len(self.point) == 1:
            desc = "v\\mapsto "
            desc += f"v'({','.join([_to_tex(i, True) for i in self.point])})"
            return desc, []
        desc = "v\\mapsto"
        desc += f"\\nabla {{v}}({','.join([_to_tex(i, True) for i in self.point])})"
        desc += "\\cdot\\left(\\begin{array}{c}"
        desc += "\\\\".join([_to_tex(i) for i in self.dir])
        desc += "\\end{array}\\right)"
        return desc, []

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Point evaluation of directional derivative"


class PointNormalDerivativeEvaluation(PointDirectionalDerivativeEvaluation):
    """A point evaluation of a normal derivative."""

    def __init__(
        self,
        reference: Reference,
        point_in: FunctionInput,
        edge: Reference,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            edge: The edge of the cell
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        assert isinstance(edge, Interval)
        super().__init__(reference, point_in, edge.normal(), entity=entity, mapping=mapping)
        self.reference = edge

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 1

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        desc = "v\\mapsto"
        desc += "\\nabla {v}(" + ",".join([_to_tex(i, True) for i in self.dof_point()]) + ")"
        entity_n = self.entity_number()
        desc += "\\cdot\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}"
        return desc, [
            "\\(\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}\\) is the normal to facet {entity_n}"
        ]

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Point evaluation of normal derivative"


class PointComponentSecondDerivativeEvaluation(BaseFunctional):
    """A point evaluation of a component of a second derivative."""

    def __init__(
        self,
        reference: Reference,
        point_in: FunctionInput,
        component: typing.Tuple[int, int],
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The point
            component: The compenent of the derivative
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.component = component

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 2

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        return function.jacobian_component(self.component).subs(x, self.point)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        desc = "v\\mapsto"
        desc += "\\frac{\\partial^2v}{"
        for c in self.component:
            desc += "\\partial " + "xyz"[c]
        desc += "}(" + ",".join([_to_tex(i, True) for i in self.dof_point()]) + ")"
        return desc, []

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Point evaluation of Hessian component"


class PointInnerProduct(BaseFunctional):
    """An evaluation of an inner product at a point."""

    def __init__(
        self,
        reference: Reference,
        point_in: FunctionInput,
        lvec: FunctionInput,
        rvec: FunctionInput,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The points
            lvec: The vector to multiply on the left
            rvec: The vector to multiply on the right
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.lvec = parse_function_input(lvec)
        self.rvec = parse_function_input(rvec)
        assert self.lvec.is_vector
        assert self.rvec.is_vector

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 0

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        v = function.subs(x, self.point)

        return self.lvec.dot(v @ self.rvec)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        if self.rvec != self.lvec:
            return None
        lv = self.lvec.as_sympy()
        assert isinstance(lv, tuple)
        return lv

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        assert isinstance(self.lvec, VectorFunction)
        assert isinstance(self.rvec, VectorFunction)
        desc = "\\mathbf{V}\\mapsto"
        desc += "\\left(\\begin{array}{c}"
        desc += "\\\\".join([_to_tex(i) for i in self.lvec])
        desc += "\\end{array}\\right)^{\\text{t}}"
        desc += "\\mathbf{V}(" + ",".join([_to_tex(i, True) for i in self.dof_point()]) + ")"
        desc += "\\left(\\begin{array}{c}"
        desc += "\\\\".join([_to_tex(i) for i in self.rvec])
        desc += "\\end{array}\\right)"
        return desc, []

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        return [[float(i) for i in self.dof_point()]], [
            [[float(i * j)]] for i in self.lvec for j in self.rvec
        ]

    name = "Point inner product"


class DotPointEvaluation(BaseFunctional):
    """A point evaluation in a given direction."""

    def __init__(
        self,
        reference: Reference,
        point_in: FunctionInput,
        vector_in: FunctionInput,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The points
            vector_in: The vector to dot with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector
        self.vector = parse_function_input(vector_in)

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 0

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        v1 = function.subs(x, self.point)
        v2 = self.vector.subs(x, self.point)
        return v1.dot(v2)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        v = self.vector.as_sympy()
        if isinstance(v, tuple):
            return v
        return None

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        desc = "\\boldsymbol{v}\\mapsto"
        desc += "\\boldsymbol{v}(" + ",".join([_to_tex(i, True) for i in self.dof_point()]) + ")"
        if self.vector.is_vector:
            desc += "\\cdot\\left(\\begin{array}{c}"
            assert hasattr(self.vector, "__iter__")
            desc += "\\\\".join([_to_tex(i) for i in self.vector])
            desc += "\\end{array}\\right)"
        elif self.vector != 1:
            desc += f"\\cdot{_to_tex(self.vector)}"
        return desc, []

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        if self.vector.is_vector:
            return [[float(i) for i in self.dof_point()]], [[[float(i)]] for i in self.vector]
        elif self.vector.is_matrix:
            return [[float(i) for i in self.dof_point()]], [
                [[float(self.vector[i][j])]]
                for i in range(self.vector.shape[0])
                for j in range(self.vector.shape[1])
            ]
        else:
            raise NotImplementedError()

    name = "Dot point evaluation"


class PointDivergenceEvaluation(BaseFunctional):
    """A point evaluation of the divergence."""

    def __init__(
        self,
        reference: Reference,
        point_in: FunctionInput,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            point_in: The points
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.point = parse_function_input(point_in)
        assert self.point.is_vector

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 1

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        out = ScalarFunction(0)
        if isinstance(function, PiecewiseFunction):
            pt = self.point.as_sympy()
            assert isinstance(pt, tuple)
            function = function.get_piece(pt)
        assert isinstance(function, VectorFunction)
        for f, i in zip(function, x):
            out += f.diff(i)
        return out.subs(x, self.point)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        pt = self.point.as_sympy()
        assert isinstance(pt, tuple)
        return pt

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        return None

    def adjusted_dof_point(self) -> PointType:
        """Get the adjusted position of the DOF in the cell for plotting.

        Returns:
            The point
        """
        return self.dof_point()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        desc = "\\boldsymbol{v}\\mapsto"
        desc += "\\nabla a\\cdot\\boldsymbol{v}"
        desc += "(" + ",".join([_to_tex(i, True) for i in self.dof_point()]) + ")"
        return desc, []

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Point evaluation of divergence"


class IntegralAgainst(BaseFunctional):
    """An integral against a function."""

    f: Function

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f_in: The function to multiply with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.integral_domain = reference.sub_entity(*entity)

        f = parse_function_input(f_in) / reference.jacobian()
        f = f.subs(x, t)

        if f.is_vector:
            assert len(f) == self.integral_domain.tdim
            self.f = mappings.contravariant(
                f,
                self.integral_domain.get_map_to_self(),
                self.integral_domain.get_inverse_map_to_self(),
            )
        elif f.is_matrix:
            assert f.shape[0] == self.integral_domain.tdim
            assert f.shape[1] == self.integral_domain.tdim
            self.f = mappings.double_contravariant(
                f,
                self.integral_domain.get_map_to_self(),
                self.integral_domain.get_inverse_map_to_self(),
            )
        else:
            self.f = f

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 0

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        return tuple(sympy.Rational(sum(i), len(i)) for i in zip(*self.integral_domain.vertices))

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        v1 = function.subs(x, point)
        integrand = self.dot(v1)
        return integrand.integral(self.integral_domain)

    def dot(self, function: Function) -> ScalarFunction:
        """Dot a function with the moment function.

        Args:
            function: The function

        Returns:
            The product of the function and the moment function
        """
        return function.dot(self.f)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        if self.f.is_vector:
            desc = "\\mathbf{v}\\mapsto"
            desc += f"\\displaystyle\\int_{{{entity}}}"
            desc += _to_tex(self.f, True) + "\\cdot"
            desc += "\\mathbf{v}"
        else:
            desc = "\\mathbf{v}\\mapsto"
            desc += f"\\displaystyle\\int_{{{entity}}}"
            if self.f != 1:
                desc += "(" + _to_tex(self.f, True) + ")"
            desc += "v"
        return desc, [entity_def]

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        return discrete_integral_moment(self.integral_domain, self.f, poly_degree)

    name = "Integral against"


class IntegralOfDivergenceAgainst(BaseFunctional):
    """An integral of the divergence against a function."""

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f_in: The function to multiply with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.integral_domain = reference.sub_entity(*entity)

        f = parse_function_input(f_in)
        self.f = f.subs(x, t)

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 1

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        return tuple(sympy.Rational(sum(i), len(i)) for i in zip(*self.integral_domain.vertices))

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        v1 = function.div().subs(x, point)
        integrand = self.dot(v1)
        return integrand.integral(self.integral_domain)

    def dot(self, function: ScalarFunction) -> ScalarFunction:
        """Dot a function with the moment function.

        Args:
            function: The function

        Returns:
            The product of the function and the moment function
        """
        return function * self.f

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\mathbf{v}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += "\\nabla\\cdot\\mathbf{v}"
        return desc, [entity_def]

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Integral of divergence against"


class IntegralOfDirectionalMultiderivative(BaseFunctional):
    """An integral of a directional derivative of a scalar function."""

    def __init__(
        self,
        reference: Reference,
        directions: SetOfPoints,
        orders: typing.Tuple[int, ...],
        entity: typing.Tuple[int, int],
        scale: int = 1,
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            directions: The directions of the derivatives
            orders: The orders of the derivatives
            entity: The entity the functional is associated with
            scale: The scale factor
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, entity, mapping)
        self.integral_domain = reference.sub_entity(*entity)
        self.directions = directions
        self.orders = orders
        self.scale = scale

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return len(self.directions)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        return tuple(sympy.Rational(sum(i), len(i)) for i in zip(*self.integral_domain.vertices))

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        for dir, o in zip(self.directions, self.orders):
            for i in range(o):
                function = function.grad(len(dir)).dot(VectorFunction(dir))
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        integrand = ScalarFunction(self.scale) * function.subs(x, point)
        return integrand.integral(self.integral_domain)

    def perform_mapping(
        self, fs: typing.List[Function], map: PointType, inverse_map: PointType
    ) -> typing.List[Function]:
        """Map functions to a cell.

        Args:
            fs: functions
            map: Map from the reference cell to a physical cell
            inverse_map: Map to the reference cell from a physical cell

        Returns:
            Mapped functions
        """
        raise mappings.MappingNotImplemented()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\mathbf{V}\\mapsto"
        desc += "\\displaystyle "
        if self.scale != 1:
            desc += _to_tex(self.scale)
        desc += f"\\int_{{{entity}}}"
        for order, dir in zip(self.orders, self.directions):
            if order > 0:
                desc += "\\frac{\\partial"
                if order > 1:
                    desc += f"^{{{order}}}"
                desc += "}{"
                desc += "\\partial" + _to_tex(dir)
                if order > 1:
                    desc += f"^{{{order}}}"
                desc += "}"
        desc += "v"
        return desc, [entity_def]

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Integral of a directional derivative"


class IntegralMoment(BaseFunctional):
    """An integral moment."""

    f: Function

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        dof: BaseFunctional,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
        map_function: bool = True,
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            integral_domain: The domain of the integral
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
            map_function: Should the function be mapped?
        """
        super().__init__(reference, entity, mapping)
        self.integral_domain = reference.sub_entity(*entity)
        self.dof = dof

        id_def = reference.default_reference().sub_entity(*entity)

        f = parse_function_input(f_in)
        self.f = f.subs(x, t)

        # Map from reference entity to entity
        if id_def.default_reference() != id_def:
            if self.f.is_vector and len(self.f) != reference.gdim:
                self.f = mappings.contravariant(
                    self.f, id_def.get_map_to_self(), id_def.get_inverse_map_to_self()
                )
            elif self.f.is_matrix:
                assert self.f.shape[0] == id_def.tdim
                assert self.f.shape[1] == id_def.tdim
                self.f = mappings.double_contravariant(
                    self.f, id_def.get_map_to_self(), id_def.get_inverse_map_to_self()
                )

        # Map from default reference to reference
        if map_function and reference != reference.default_reference():
            if mapping is None or not hasattr(mappings, f"{mapping}_inverse_transpose"):
                raise NonDefaultReferenceError()
            mf = getattr(mappings, f"{mapping}_inverse_transpose")
            self.f = mf(
                self.f,
                reference.get_map_to_self(),
                reference.get_inverse_map_to_self(),
                substitute=False,
            )
            self.f *= id_def.volume() / self.integral_domain.volume()

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 0

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k

        v1 = function.subs(x, point)
        integrand = self.dot(v1)
        return integrand.integral(self.integral_domain)

    def dot(self, function: Function) -> ScalarFunction:
        """Dot a function with the moment function.

        Args:
            function: The function

        Returns:
            The product of the function and the moment function
        """
        return function.dot(self.f)

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            The point
        """
        p = self.dof.dof_point()
        return tuple(
            o + sum(self.integral_domain.axes[j][i] * c for j, c in enumerate(p))
            for i, o in enumerate(self.integral_domain.origin)
        )

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        p = self.dof.dof_direction()
        if p is None:
            if self.f.is_vector:
                p = (self.f.subs(t, self.dof_point())).as_sympy()
                assert isinstance(p, tuple)
                return p
            return None
        vp = VectorFunction(p)
        out = []
        for i in range(self.integral_domain.gdim):
            entry = vp.dot(VectorFunction([a[i] for a in self.integral_domain.axes])).as_sympy()
            assert isinstance(entry, sympy.core.expr.Expr)
            out.append(entry)
        return tuple(out)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        if self.f.is_vector:
            if len(self.f) in [self.reference.tdim, self.integral_domain.tdim]:
                desc = "\\boldsymbol{v}\\mapsto"
                desc += f"\\displaystyle\\int_{{{entity}}}"
                desc += "\\boldsymbol{v}\\cdot"
                desc += "\\left(\\begin{array}{c}"
                assert hasattr(self.f, "__iter__")
                desc += "\\\\".join([_to_tex(i) for i in self.f])
                desc += "\\end{array}\\right)"
            else:
                if len(self.f) == self.integral_domain.tdim**2:
                    size = self.integral_domain.tdim
                elif len(self.f) == self.reference.tdim**2:
                    size = self.reference.tdim
                else:
                    raise NotImplementedError()
                desc = "\\mathbf{V}\\mapsto"
                desc += f"\\displaystyle\\int_{{{entity}}}"
                desc += "\\mathbf{V}:"
                desc += "\\left(\\begin{array}{" + "c" * size + "}"
                desc += "\\\\".join(
                    [
                        "&".join([_to_tex(self.f[i]) for i in range(size * row, size * (row + 1))])
                        for row in range(size)
                    ]
                )
                desc += "\\end{array}\\right)"
        else:
            desc = "v\\mapsto"
            desc += f"\\displaystyle\\int_{{{entity}}}"
            if self.f != 1:
                desc += "(" + _to_tex(self.f) + ")"
            desc += "v"
        return desc, [entity_def]

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        return discrete_integral_moment(self.integral_domain, self.f, poly_degree)

    name = "Integral moment"


class DerivativeIntegralMoment(IntegralMoment):
    """An integral moment of the derivative of a scalar function."""

    def __init__(
        self,
        reference: Reference,
        f: FunctionInput,
        dot_with_in: FunctionInput,
        dof: BaseFunctional,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f: The function to multiply with
            dot_with_in: The vector to take the dot product with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        super().__init__(reference, f, dof, entity=entity, mapping=mapping)
        self.dot_with = parse_function_input(dot_with_in)

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 1

    def dot(self, function: Function) -> ScalarFunction:
        """Dot a function with the moment function.

        Args:
            function: The function

        Returns:
            The product of the function and the moment function
        """
        assert function.is_vector
        assert self.f.is_scalar
        return function.dot(self.dot_with) * self.f

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        dw = self.dot_with.as_sympy()
        assert isinstance(dw, tuple)
        return dw

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        integrand = self.dot(function.grad(self.integral_domain.gdim).subs(x, point))
        value = integrand.integral(self.integral_domain)
        return value

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Derivative integral moment"


class DivergenceIntegralMoment(IntegralMoment):
    """An integral moment of the divergence of a vector function."""

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        dof: BaseFunctional,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        assert f.is_scalar
        super().__init__(reference, f, dof, entity=entity, mapping=mapping)

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 1

    def _eval_symbolic(self, function: Function) -> Function:
        """Apply to the functional to a function.

        Args:
            function: The function

        Returns:
            The value of the functional for the function
        """
        point = [i for i in self.integral_domain.origin]
        for i, a in enumerate(zip(*self.integral_domain.axes)):
            for j, k in zip(a, t):
                point[i] += j * k
        assert function.is_vector
        integrand = self.dot(function.div().subs(x, point))
        return integrand.integral(self.integral_domain)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{v}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += "\\nabla\\cdot\\boldsymbol{v}"
        return desc, [entity_def]

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Integral moment of divergence"


class TangentIntegralMoment(IntegralMoment):
    """An integral moment in the tangential direction."""

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        dof: BaseFunctional,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "covariant",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        self._scalar_f = parse_function_input(f_in).subs(x, t)
        integral_domain = reference.sub_entity(*entity)
        assert self._scalar_f.is_scalar
        super().__init__(
            reference,
            tuple(self._scalar_f * i for i in integral_domain.tangent()),
            dof,
            entity=entity,
            mapping=mapping,
            map_function=False,
        )

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        return self.integral_domain.tangent()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_n = self.entity_number()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{v}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        desc += "\\boldsymbol{v}\\cdot"
        if self.f != 1:
            desc += "(" + _to_tex(self._scalar_f) + ")"
        desc += "\\hat{\\boldsymbol{t}}" + f"_{{{entity_n}}}"
        return desc, [
            entity_def,
            f"\\(\\hat{{\\boldsymbol{{t}}}}_{{{entity_n}}}\\) is the tangent to edge {entity_n}",
        ]

    name = "Tangential integral moment"


class NormalIntegralMoment(IntegralMoment):
    """An integral moment in the normal direction."""

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        dof: BaseFunctional,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "contravariant",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        self._scalar_f = parse_function_input(f_in).subs(x, t)
        assert self._scalar_f.is_scalar
        integral_domain = reference.sub_entity(*entity)
        super().__init__(
            reference,
            tuple(self._scalar_f * i for i in integral_domain.normal()),
            dof,
            entity=entity,
            mapping=mapping,
            map_function=False,
        )

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        return self.integral_domain.normal()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_n = self.entity_number()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{v}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        desc += "\\boldsymbol{v}\\cdot"
        if self.f != 1:
            desc += "(" + _to_tex(self._scalar_f, True) + ")"
        desc += "\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}"
        return desc, [
            entity_def,
            f"\\(\\hat{{\\boldsymbol{{n}}}}_{{{entity_n}}}\\) is the normal to facet {entity_n}",
        ]

    name = "Normal integral moment"


class TraceIntegralMoment(IntegralMoment):
    """An integral moment of the trace of a matrix."""

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        dof: BaseFunctional,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "contravariant",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        scalar_f = parse_function_input(f_in).subs(x, t)
        assert scalar_f.is_scalar
        super().__init__(
            reference,
            scalar_f,
            dof,
            entity=entity,
            mapping=mapping,
            map_function=False,
        )

    def dot(self, function: Function) -> ScalarFunction:
        """Take the product of a function with the trace of the matrix.

        Args:
            function: The function

        Returns:
            The inner product of the function and the matrix's trace
        """
        assert function.is_matrix and function.shape[0] == function.shape[1]
        trace = sum(function[i, i] for i in range(function.shape[0]))
        return trace * self.f * self.integral_domain.jacobian()

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{V}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        desc += "\\operatorname{tr}(\\boldsymbol{V})"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        return desc, [entity_def]

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Trace integral moment"


class NormalDerivativeIntegralMoment(DerivativeIntegralMoment):
    """An integral moment in the normal direction."""

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        dof: BaseFunctional,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        assert f.is_scalar
        integral_domain = reference.sub_entity(*entity)
        super().__init__(
            reference, f, integral_domain.normal(), dof, entity=entity, mapping=mapping
        )

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_n = self.entity_number()
        entity_def = self.entity_definition()
        desc = "v\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += "\\frac{\\partial v}"
        desc += "{\\partial\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}" + "}"
        return desc, [
            entity_def,
            f"\\(\\hat{{\\boldsymbol{{n}}}}_{{{entity_n}}}\\) is the normal to facet {entity_n}",
        ]

    @property
    def nderivs(self) -> int:
        """Number of derivatives."""
        return 1

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        raise NotImplementedError()

    name = "Normal derivative integral moment"


class InnerProductIntegralMoment(IntegralMoment):
    """An integral moment of the inner product with a vector."""

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        inner_with_left_in: FunctionInput,
        inner_with_right_in: FunctionInput,
        dof: BaseFunctional,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "identity",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            inner_with_left_in: The vector to multiply on the left
            inner_with_right_in: The vector to multiply on the right
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        inner_with_left = parse_function_input(inner_with_left_in)
        inner_with_right = parse_function_input(inner_with_right_in)
        assert f.is_scalar
        assert inner_with_left.is_vector
        assert inner_with_right.is_vector

        super().__init__(reference, f, dof, entity=entity, mapping=mapping)
        self.inner_with_left = inner_with_left
        self.inner_with_right = inner_with_right

    def dot(self, function: Function) -> ScalarFunction:
        """Take the inner product of a function with the moment direction.

        Args:
            function: The function

        Returns:
            The inner product of the function and the moment direction
        """
        assert function.is_matrix
        return (
            self.inner_with_left.dot(function @ self.inner_with_right)
            * self.f
            * self.integral_domain.jacobian()
        )

    def dof_direction(self) -> typing.Union[PointType, None]:
        """Get the direction of the DOF.

        Returns:
            The direction
        """
        if self.inner_with_left != self.inner_with_right:
            return None
        il = self.inner_with_left.as_sympy()
        assert isinstance(il, tuple)
        return il

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_def = self.entity_definition()
        desc = "\\boldsymbol{V}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += _to_tex(self.inner_with_left) + "^{\\text{t}}"
        desc += "\\boldsymbol{V}"
        desc += _to_tex(self.inner_with_right)
        return desc, [entity_def]

    def discrete(self, poly_degree: int) -> DiscreteDof:
        """Get points and weights that define this DOF discretely.

        Args:
            poly_degree: The polynomial degree of the element. This may be used to decide which
                    degree quadrature rule to use

        Returns:
            Points (a list of lists whose indices are [point_index][dimension]) and weights (a list of list of lists whose indices are [dimension][point_index][derivative])
        """
        dot_with = MatrixFunction(
            [[self.f * i * j for j in self.inner_with_right] for i in self.inner_with_left]
        )

        return discrete_integral_moment(self.integral_domain, dot_with, poly_degree)

    name = "Inner product integral moment"


class NormalInnerProductIntegralMoment(InnerProductIntegralMoment):
    """An integral moment of the inner product with the normal direction."""

    def __init__(
        self,
        reference: Reference,
        f_in: FunctionInput,
        dof: BaseFunctional,
        entity: typing.Tuple[int, int],
        mapping: typing.Union[str, None] = "double_contravariant",
    ):
        """Create the functional.

        Args:
            reference: The reference cell
            f_in: The function to multiply with
            dof: The DOF in a moment space that the function is associated with
            entity: The entity the functional is associated with
            mapping: The type of mappping from the reference cell to a physical cell
        """
        f = parse_function_input(f_in)
        assert f.is_scalar
        integral_domain = reference.sub_entity(*entity)
        super().__init__(
            reference,
            f,
            integral_domain.normal(),
            integral_domain.normal(),
            dof,
            entity=entity,
            mapping=mapping,
        )

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            Representation of the functional as TeX, and list of terms involved
        """
        entity = self.entity_tex()
        entity_n = self.entity_number()
        entity_def = self.entity_definition()
        desc = "\\mathbf{V}\\mapsto"
        desc += f"\\displaystyle\\int_{{{entity}}}"
        if self.f != 1:
            desc += "(" + _to_tex(self.f, True) + ")"
        desc += f"|{{{entity}}}|"
        desc += "\\hat{\\boldsymbol{n}}^{\\text{t}}" + f"_{{{entity_n}}}"
        desc += "\\mathbf{V}"
        desc += "\\hat{\\boldsymbol{n}}" + f"_{{{entity_n}}}"
        return desc, [
            entity_def,
            f"\\(\\hat{{\\boldsymbol{{n}}}}_{{{entity_n}}}\\) is the normal to facet {entity_n}",
        ]

    name = "Normal inner product integral moment"


# Types
ListOfFunctionals = typing.List[BaseFunctional]



"""Hsieh-Clough-Tocher elements on simplices.

This element's definition appears in https://doi.org/10.2307/2006147
(Ciarlet, 1978)
"""

import typing

import sympy

from symfem.finite_element import CiarletElement
from symfem.functionals import (
    DerivativePointEvaluation,
    ListOfFunctionals,
    PointEvaluation,
    PointNormalDerivativeEvaluation,
)
from symfem.functions import FunctionInput, ScalarFunction
from symfem.piecewise_functions import PiecewiseFunction
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import x

__all__ = ["HsiehCloughTocher"]


class HsiehCloughTocher(CiarletElement):
    """Hsieh-Clough-Tocher finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 3
        assert reference.name == "triangle"
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, vs, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (0, 1), entity=(0, v_n)))
        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            dofs.append(
                PointNormalDerivativeEvaluation(
                    reference, sub_ref.midpoint(), sub_ref, entity=(1, e_n)
                )
            )

        mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*reference.vertices))

        subs = [
            (reference.vertices[0], reference.vertices[1], mid),
            (reference.vertices[1], reference.vertices[2], mid),
            (reference.vertices[2], reference.vertices[0], mid),
        ]

        piece_list = [
            tuple(ScalarFunction(p) for _ in range(3))
            for p in [
                1,
                x[0],
                x[1],
                x[0] ** 2,
                x[0] * x[1],
                x[1] ** 2,
                x[0] ** 3,
                x[0] ** 2 * x[1],
                x[0] * x[1] ** 2,
                x[1] ** 3,
            ]
        ]
        piece_list.append(
            (
                ScalarFunction(
                    -23 * x[0] ** 3 + 24 * x[0] ** 2 * x[1] - 12 * x[0] * x[1] ** 2 + 36 * x[1] ** 2
                ),
                ScalarFunction(
                    -28 * x[0] ** 3
                    + 12 * x[0] ** 2 * x[1]
                    + 9 * x[0] ** 2
                    - 3 * x[0]
                    + 32 * x[1] ** 3
                    + 12 * x[1]
                    - 1
                ),
                ScalarFunction(
                    -15 * x[0] ** 2
                    - 33 * x[0] * x[1] ** 2
                    + 30 * x[0] * x[1]
                    + 22 * x[1] ** 3
                    + 21 * x[1] ** 2
                ),
            )
        )
        piece_list.append(
            (
                ScalarFunction(
                    22 * x[0] ** 3
                    - 21 * x[0] ** 2 * x[1]
                    - 12 * x[0] * x[1] ** 2
                    + 30 * x[0] * x[1]
                    - 24 * x[1] ** 2
                ),
                ScalarFunction(
                    32 * x[0] ** 3
                    + 12 * x[0] ** 2 * x[1]
                    - 21 * x[0] ** 2
                    + 12 * x[0]
                    - 28 * x[1] ** 3
                    - 3 * x[1]
                    - 1
                ),
                ScalarFunction(
                    15 * x[0] ** 2 + 12 * x[0] * x[1] ** 2 - 23 * x[1] ** 3 - 9 * x[1] ** 2
                ),
            )
        )

        poly: typing.List[FunctionInput] = []
        poly += [PiecewiseFunction({i: j for i, j in zip(subs, p)}, 2) for p in piece_list]
        poly = reference.map_polyset_from_default(poly)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    @property
    def lagrange_subdegree(self) -> int:
        raise NotImplementedError()

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        raise NotImplementedError()

    @property
    def polynomial_subdegree(self) -> int:
        raise NotImplementedError()

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        raise NotImplementedError()

    names = ["Hsieh-Clough-Tocher", "Clough-Tocher", "HCT", "CT"]
    references = ["triangle"]
    min_order = 3
    max_order = 3
    # continuity = "C1"
    continuity = "C0"
    value_type = "scalar macro"
    last_updated = "2023.06"



"""Guzman-Neilan elements on simplices.

This element's definition appears in https://doi.org/10.1137/17M1153467
(Guzman and Neilan, 2018)
"""

import typing

import sympy

from symfem.elements.bdm import BDM
from symfem.elements.bernardi_raugel import BernardiRaugel
from symfem.elements.lagrange import Lagrange, VectorLagrange
from symfem.finite_element import CiarletElement
from symfem.functionals import DotPointEvaluation, ListOfFunctionals, NormalIntegralMoment
from symfem.functions import Function, FunctionInput, ScalarFunction, VectorFunction
from symfem.geometry import SetOfPoints, SetOfPointsInput
from symfem.moments import make_integral_moment_dofs
from symfem.piecewise_functions import PiecewiseFunction
from symfem.references import NonDefaultReferenceError, Reference
from symfem.symbols import t, x

__all__ = ["GuzmanNeilanFirstKind", "GuzmanNeilanSecondKind", "make_piecewise_lagrange"]


def poly(reference: Reference, k: int) -> typing.List[Function]:
    """Generate the P^perp polynomial set."""
    if k < 2:
        if reference.name == "triangle":
            return [
                VectorFunction(
                    [x[0] ** i0 * x[1] ** i1 if d2 == d else 0 for d2 in range(reference.tdim)]
                )
                for d in range(reference.tdim)
                for i0 in range(k + 1)
                for i1 in range(k + 1 - i0)
            ]
        else:
            assert reference.name == "tetrahedron"
            return [
                VectorFunction(
                    [
                        x[0] ** i0 * x[1] ** i1 * x[2] ** i2 if d2 == d else 0
                        for d2 in range(reference.tdim)
                    ]
                )
                for d in range(reference.tdim)
                for i0 in range(k + 1)
                for i1 in range(k + 1 - i0)
                for i2 in range(k + 1 - i0 - i1)
            ]

    e = BDM(reference, k)
    return [i for (i, j) in zip(e.get_basis_functions(), e.dofs) if j.entity != (reference.tdim, 0)]


def get_sub_cells(reference):
    """Get sub cells."""
    mid = reference.midpoint()
    if reference.name == "triangle":
        return [
            (reference.vertices[0], reference.vertices[1], mid),
            (reference.vertices[0], reference.vertices[2], mid),
            (reference.vertices[1], reference.vertices[2], mid),
        ]
    if reference.name == "tetrahedron":
        return [
            (reference.vertices[0], reference.vertices[1], reference.vertices[2], mid),
            (reference.vertices[0], reference.vertices[1], reference.vertices[3], mid),
            (reference.vertices[0], reference.vertices[2], reference.vertices[3], mid),
            (reference.vertices[1], reference.vertices[2], reference.vertices[3], mid),
        ]


def bubbles(reference: Reference) -> typing.List[VectorFunction]:
    """Generate divergence-free bubbles."""
    br_bubbles = []
    p1 = Lagrange(reference, 1, variant="equispaced")
    for i in range(reference.sub_entity_count(reference.tdim - 1)):
        sub_e = reference.sub_entity(reference.tdim - 1, i)
        bubble = ScalarFunction(1)
        for j in reference.sub_entities(reference.tdim - 1)[i]:
            bubble *= p1.get_basis_function(j)
        br_bubbles.append(VectorFunction(tuple(bubble * j for j in sub_e.normal())))

    sub_cells = get_sub_cells(reference)
    xx, yy, zz = x

    if reference.name == "triangle":
        terms = [1, xx, yy]

        lamb = PiecewiseFunction(
            {sub_cells[0]: 3 * x[1], sub_cells[1]: 3 * x[0], sub_cells[2]: 3 * (1 - x[0] - x[1])}, 2
        )

    if reference.name == "tetrahedron":
        terms = [1, xx, yy, zz, xx**2, yy**2, zz**2, xx * yy, xx * zz, yy * zz]

        lamb = PiecewiseFunction(
            {
                sub_cells[0]: 4 * x[2],
                sub_cells[1]: 4 * x[1],
                sub_cells[2]: 4 * x[0],
                sub_cells[3]: 4 * (1 - x[0] - x[1] - x[2]),
            },
            3,
        )

    sub_basis = [
        p * lamb**j
        for j in range(1, reference.tdim + 1)
        for p in poly(reference, reference.tdim - j)
    ]

    bubbles = []

    for f in br_bubbles:
        assert isinstance(f, VectorFunction)
        integrand = f.div().subs(x, t)
        fun_s = (f.div() - integrand.integral(reference) / reference.volume()).as_sympy()

        assert isinstance(fun_s, sympy.core.expr.Expr)
        fun = {x[0] ** i[0] * x[1] ** i[1] * x[2] ** i[2]: j for i, j in fun_s.as_poly(x).terms()}

        for term in fun:
            assert term in terms
        aim = [fun[term] if term in fun else 0 for term in terms] * (reference.tdim + 1)

        mat: typing.List[typing.List[ScalarFunction]] = [[] for t in terms for p in lamb.pieces]
        for b in sub_basis:
            assert isinstance(b, PiecewiseFunction)
            i = 0
            for p in b.pieces.values():
                assert isinstance(p, VectorFunction)
                d_s = p.div().as_sympy()
                assert isinstance(d_s, sympy.core.expr.Expr)
                d = d_s.expand().as_coefficients_dict()
                for term in d:
                    assert term == 0 or term in terms
                for term in terms:
                    if i < len(mat):
                        mat[i].append(d[term] if term in d else 0)  # type: ignore
                    i += 1

        s_mat = sympy.Matrix(mat)
        solution = (s_mat.T @ s_mat).inv() @ s_mat.T @ sympy.Matrix(aim)
        assert s_mat @ solution == sympy.Matrix(aim)
        coeffs = list(solution)

        for coeff, basis_f in zip(coeffs, sub_basis):
            f -= coeff * basis_f
        bubbles.append(f)

    return bubbles


class GuzmanNeilanFirstKind(CiarletElement):
    """Guzman-Neilan finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        sub_cells = get_sub_cells(reference)

        lagrange = VectorLagrange(reference, order)
        poly = [
            PiecewiseFunction({i: p for i in sub_cells}, reference.tdim)
            for p in lagrange.get_polynomial_basis()
        ] + bubbles(reference)

        br = BernardiRaugel(reference, order)
        if order == 1:
            dofs = br.dofs
        else:
            assert order == 2 and reference.name == "tetrahedron"
            dofs = br.dofs[:-3]

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)  # type: ignore

    @property
    def lagrange_subdegree(self) -> int:
        raise NotImplementedError()

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        raise NotImplementedError()

    @property
    def polynomial_subdegree(self) -> int:
        raise NotImplementedError()

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        raise NotImplementedError()

    names = ["Guzman-Neilan first kind", "Guzman-Neilan"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    max_order = {"triangle": 1, "tetrahedron": 2}
    continuity = "L2"
    value_type = "vector macro"
    last_updated = "2024.10.5"


class GuzmanNeilanSecondKind(CiarletElement):
    """Guzman-Neilan second kind finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()

        sub_cells = get_sub_cells(reference)

        poly = make_piecewise_lagrange(sub_cells, reference.name, order) + bubbles(reference)

        dofs: ListOfFunctionals = []

        tdim = reference.tdim

        # Evaluation at vertices
        for n in range(reference.sub_entity_count(0)):
            vertex = reference.sub_entity(0, n)
            v = vertex.vertices[0]
            for i in range(tdim):
                direction = tuple(1 if i == j else 0 for j in range(tdim))
                dofs.append(
                    DotPointEvaluation(
                        reference,
                        v,
                        direction,
                        entity=(0, n),
                        mapping="identity",
                    )
                )

        if order == 2:
            assert reference.name == "tetrahedron"

            # Midpoints of edges
            for n in range(reference.sub_entity_count(1)):
                edge = reference.sub_entity(1, n)
                for i in range(tdim):
                    direction = tuple(1 if i == j else 0 for j in range(tdim))
                    dofs.append(
                        DotPointEvaluation(
                            reference,
                            edge.midpoint(),
                            direction,
                            entity=(1, n),
                            mapping="identity",
                        )
                    )

            # Interior edges
            for v in reference.vertices:
                p = tuple((i + sympy.Rational(1, 4)) / 2 for i in v)
                for d in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]:
                    dofs.append(
                        DotPointEvaluation(reference, p, d, entity=(3, 0), mapping="identity")
                    )

        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Lagrange, 0, "contravariant"),
        )

        mid = reference.midpoint()
        for i in range(tdim):
            direction = tuple(1 if i == j else 0 for j in range(tdim))
            dofs.append(
                DotPointEvaluation(reference, mid, direction, entity=(tdim, 0), mapping="identity")
            )

        super().__init__(reference, order, poly, dofs, tdim, tdim)  # type: ignore

    @property
    def lagrange_subdegree(self) -> int:
        raise NotImplementedError()

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        raise NotImplementedError()

    @property
    def polynomial_subdegree(self) -> int:
        raise NotImplementedError()

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        raise NotImplementedError()

    names = ["Guzman-Neilan second kind"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    max_order = {"triangle": 1, "tetrahedron": 2}
    continuity = "L2"
    value_type = "vector macro"
    last_updated = "2024.10.3"


def make_piecewise_lagrange(
    sub_cells: typing.List[SetOfPoints],
    cell_name,
    order: int,
    zero_on_boundary: bool = False,
    zero_at_centre: bool = False,
) -> typing.List[PiecewiseFunction]:
    """Make the basis functions of a piecewise Lagrange space.

    Args:
        sub_cells: A list of vertices of sub cells
        cell_name: The cell type of the sub cells
        order: The polynomial order
        zero_in_boundary: Should the functions be zero on the boundary?
        zero_at_centre: Should the functions be zero at the centre?

    Returns:
        The basis functions
    """
    from symfem import create_reference

    lagrange_space = VectorLagrange(create_reference(cell_name), order)
    lagrange_bases: typing.List[typing.List[VectorFunction]] = []
    for c in sub_cells:
        row: typing.List[VectorFunction] = []
        c_basis = lagrange_space.map_to_cell(c)
        for cb in c_basis:
            assert isinstance(cb, VectorFunction)
            row.append(cb)
        lagrange_bases.append(row)

    basis_dofs: typing.List[typing.Tuple[int, ...]] = []
    zero: typing.Tuple[int, ...] = (0,)
    if cell_name == "triangle":
        cell_tdim = 2
        for dim, tri_entities in enumerate(
            [
                [(0, 0, -1), (1, -1, 0), (-1, 1, 1), (2, 2, 2)],
                [(0, -1, 1), (1, 1, -1), (-1, 0, 0), (2, -1, -1), (-1, 2, -1), (-1, -1, 2)],
                [(0, -1, -1), (-1, 0, -1), (-1, -1, 0)],
            ]
        ):
            nones = [-1 for i in lagrange_space.entity_dofs(dim, 0)]
            for tri_e in tri_entities:
                if dim == 0:
                    if zero_on_boundary and (0 in tri_e or 1 in tri_e):
                        continue
                    if zero_at_centre and (2 in tri_e):
                        continue
                elif dim == 1:
                    if zero_on_boundary and (2 in tri_e):
                        continue
                doflist = [nones if i == -1 else lagrange_space.entity_dofs(dim, i) for i in tri_e]
                for dofs in zip(*doflist):
                    basis_dofs.append(dofs)
        zero = (0, 0)

    elif cell_name == "tetrahedron":
        cell_tdim = 3
        for dim, tet_entities in enumerate(
            [
                [(0, 0, 0, -1), (1, 1, -1, 0), (2, -1, 1, 1), (-1, 2, 2, 2), (3, 3, 3, 3)],
                [
                    (2, -1, -1, 5),
                    (5, 5, -1, -1),
                    (4, -1, 5, -1),
                    (-1, 2, -1, 4),
                    (-1, 4, 4, -1),
                    (-1, -1, 2, 2),
                    (3, 3, 3, -1),
                    (-1, 0, 0, 0),
                    (0, -1, 1, 1),
                    (1, 1, -1, 3),
                ],
                [
                    (0, -1, -1, 2),
                    (1, -1, 2, -1),
                    (2, 2, -1, -1),
                    (3, -1, -1, -1),
                    (-1, 0, -1, 1),
                    (-1, 1, 1, -1),
                    (-1, 3, -1, -1),
                    (-1, -1, 0, 0),
                    (-1, -1, 3, -1),
                    (-1, -1, -1, 3),
                ],
                [(0, -1, -1, -1), (-1, 0, -1, -1), (-1, -1, 0, -1), (-1, -1, -1, 0)],
            ]
        ):
            nones = [-1 for i in lagrange_space.entity_dofs(dim, 0)]
            for tet_e in tet_entities:
                if dim == 0:
                    if zero_on_boundary and (0 in tet_e or 1 in tet_e or 2 in tet_e):
                        continue
                    if zero_at_centre and (3 in tet_e):
                        continue
                elif dim == 1:
                    if zero_on_boundary and (2 in tet_e or 4 in tet_e or 5 in tet_e):
                        continue
                elif dim == 2:
                    if zero_on_boundary and (3 in tet_e):
                        continue
                doflist = [nones if i == -1 else lagrange_space.entity_dofs(dim, i) for i in tet_e]
                for dofs in zip(*doflist):
                    basis_dofs.append(dofs)
        zero = (0, 0, 0)

    basis: typing.List[PiecewiseFunction] = []
    for i in basis_dofs:
        pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {i: zero for i in sub_cells}
        for c, s, j in zip(sub_cells, lagrange_bases, i):
            if j != -1:
                pieces[c] = s[j]
        basis.append(PiecewiseFunction(pieces, cell_tdim))

    return basis



"""Morley elements on simplices.

This element's definition appears in https://doi.org/10.1017/S0001925900004546
(Morley, 1968)
"""

import typing

from symfem.finite_element import CiarletElement
from symfem.functionals import ListOfFunctionals, PointEvaluation, PointNormalDerivativeEvaluation
from symfem.functions import FunctionInput
from symfem.polynomials import polynomial_set_1d
from symfem.references import NonDefaultReferenceError, Reference

__all__ = ["Morley"]


class Morley(CiarletElement):
    """Morley finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        if reference.vertices != reference.reference_vertices:
            raise NonDefaultReferenceError()
        assert order == 2
        assert reference.name == "triangle"
        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, vs, entity=(0, v_n)))
        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            midpoint = sub_ref.midpoint()
            dofs.append(
                PointNormalDerivativeEvaluation(reference, midpoint, sub_ref, entity=(1, e_n))
            )

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    @property
    def lagrange_subdegree(self) -> int:
        return self.order

    @property
    def lagrange_superdegree(self) -> typing.Optional[int]:
        return self.order

    @property
    def polynomial_subdegree(self) -> int:
        return self.order

    @property
    def polynomial_superdegree(self) -> typing.Optional[int]:
        return self.order

    names = ["Morley"]
    references = ["triangle"]
    min_order = 2
    max_order = 2
    continuity = "C1"
    value_type = "scalar"
    last_updated = "2023.05"



"""Jacobi polynomials."""

import typing
from functools import cache

from math import prod, factorial
import sympy

from symfem.functions import ScalarFunction
from symfem.symbols import x

__all__: typing.List[str] = []


def _jrc(
    a: int, b: int, n: int
) -> typing.Tuple[
    sympy.core.expr.Expr,
    sympy.core.expr.Expr,
    sympy.core.expr.Expr,
]:
    """Get the Jacobi recurrence relation coefficients.

    Args:
        a: The parameter a
        b: The parameter b
        n: The parameter n

    Returns:
        The Jacobi coefficients
    """
    if b != 0:
        raise NotImplementedError()
    return (
        sympy.Rational((a + 2 * n + 1) * (a + 2 * n + 2), 2 * (n + 1) * (a + n + 1)),
        sympy.Rational(a * a * (a + 2 * n + 1), 2 * (n + 1) * (a + n + 1) * (a + 2 * n)),
        sympy.Rational(n * (a + n) * (a + 2 * n + 2), (n + 1) * (a + n + 1) * (a + 2 * n)),
    )


def _jrc_monic(
    a: int, b: int, n: int
) -> typing.Tuple[
    sympy.core.expr.Expr,
    sympy.core.expr.Expr,
]:
    """Get the Jacobi recurrence relation coefficients for monic polynomials.

    Args:
        a: The parameter a
        b: The parameter b
        n: The parameter n

    Returns:
        The Jacobi coefficients
    """
    return (
        sympy.Rational(a**2 - b**2, (2 * n + a + b) * (2 * n + a + b + 2)),
        sympy.Rational(
            4 * n * (n + a) * (n + b) * (n + a + b),
            (2 * n + a + b + 1) * (2 * n + a + b) ** 2 * (2 * n + a + b - 1),
        ),
    )


@cache
def jacobi_polynomial_x(n: int, a: int, b: int) -> sympy.core.expr.Expr:
    """Get a Jacobi polynomial.

    Args:
        n: Polynomial degree
        a: The parameter a
        b: The parameter b
    """
    if a < 0:
        if b <= 0 and n + a >= 0:
            # See https://mathoverflow.net/questions/298661/jacobi-polynomials-with-negative-integer-parameters
            return (
                ((x[0] - 1) / 2) ** -a
                * ((x[0] + 1) / 2) ** -b
                * jacobi_polynomial(n + a + b, -a, -b)
            )
        raise NotImplementedError()
    if n == 0:
        return sympy.Integer(1)
    if n == 1:
        return (a + 1) + (a + b + 2) * (x[0] - 1) / 2
    i, j, k = _jrc(a, b, n - 1)
    return (i * x[0] + j) * jacobi_polynomial(n - 1, a, b) - k * jacobi_polynomial(n - 2, a, b)


def jacobi_polynomial(n: int, a: int, b: int, variable: sympy.Symbol = x[0]) -> ScalarFunction:
    """Get a Jacobi polynomial.

    Args:
        n: Polynomial degree
        a: The parameter a
        b: The parameter b
        variable: Variable to use
    """
    return ScalarFunction(jacobi_polynomial_x(n, a, b).subs(x[0], variable))


def choose(n: int, r: int) -> int:
    """Choose function."""
    return prod(range(n - r + 1, n + 1)) // factorial(r)


@cache
def monic_jacobi_polynomial_x(n: int, a: int, b: int) -> sympy.core.expr.Expr:
    """Get a monic Jacobi polynomial.

    Args:
        n: Polynomial degree
        a: The parameter a
        b: The parameter b
    """
    if n == 0:
        return sympy.Integer(1)
    if a + b == -n:
        return sum(
            choose(n + a, i) * choose(n + b, n - i) * (x[0] - 1) ** (n - i) * (x[0] + 1) ** i
            for i in range(n + 1)
        ) / choose(2 * n + a + b, n)
    if n == 1:
        return x[0] + sympy.Rational(2 * (a + 1), a + b + 2) - 1
    i, j = _jrc_monic(a, b, n - 1)
    return (x[0] + i) * monic_jacobi_polynomial(n - 1, a, b) - j * monic_jacobi_polynomial(
        n - 2, a, b
    )


def monic_jacobi_polynomial(
    n: int, a: int, b: int, variable: sympy.Symbol = x[0]
) -> ScalarFunction:
    """Get a monic Jacobi polynomial.

    Args:
        n: Polynomial degree
        a: The parameter a
        b: The parameter b
        variable: Variable to use
    """
    return ScalarFunction(monic_jacobi_polynomial_x(n, a, b).subs(x[0], variable))



"""Functions to map functions between cells."""

from __future__ import annotations

from typing import Callable

import sympy

from symfem.functions import Function, FunctionInput, MatrixFunction, parse_function_input
from symfem.geometry import PointType
from symfem.symbols import x

__all__ = [
    "MappingNotImplemented",
    "identity",
    "identity_inverse",
    "identity_inverse_transpose",
    "l2",
    "l2_inverse_transpose",
    "l2_inverse",
    "covariant",
    "covariant_inverse",
    "covariant_inverse_transpose",
    "contravariant",
    "contravariant_inverse",
    "contravariant_inverse_transpose",
    "double_covariant",
    "double_covariant_inverse",
    "double_contravariant",
    "double_contravariant_inverse",
    "co_contravariant",
    "co_contravariant_inverse",
    "get_mapping",
]


class MappingNotImplemented(NotImplementedError):
    """Exception thrown when a mapping is not implemented for an element."""


def identity(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Map functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    return f


def l2(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Map functions, scaling by the determinant of the jacobian.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    jdet = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map]).det().as_sympy()
    assert isinstance(jdet, sympy.core.expr.Expr)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    return f / abs(jdet)


def covariant(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Map H(curl) functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)

    assert f.is_vector

    j_inv_t = MatrixFunction([[i.diff(x[j]) for i in inverse_map] for j in range(tdim)])
    return j_inv_t @ f


def contravariant(
    f_in: FunctionInput, map: PointType, inverse_map: PointType, substitute: bool = True
) -> Function:
    """Map H(div) functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)

    assert f.is_vector

    jacobian = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map])
    jacobian /= jacobian.det()
    return jacobian @ f


def double_covariant(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> MatrixFunction:
    """Map matrix functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    assert f.is_matrix

    j_inv_t = MatrixFunction([[i.diff(x[j]) for i in inverse_map] for j in range(tdim)])
    return j_inv_t @ f @ j_inv_t.transpose()


def double_contravariant(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> MatrixFunction:
    """Map matrix functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    assert f.is_matrix

    jacobian = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map])
    jacobian /= jacobian.det()
    return jacobian @ f @ jacobian.transpose()


def co_contravariant(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> MatrixFunction:
    """Map matrix functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """

    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    assert f.is_matrix

    j_inv_t = MatrixFunction([[i.diff(x[j]) for i in inverse_map] for j in range(tdim)])
    jacobian = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map])
    jacobian /= jacobian.det()
    return j_inv_t @ f @ jacobian.transpose()


def identity_inverse_transpose(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Inverse transpose of identity().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    return f


def l2_inverse_transpose(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Inverse transpose of l2().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)
    jdet = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map]).det().as_sympy()
    assert isinstance(jdet, sympy.core.expr.Expr)
    return f * abs(jdet)


def covariant_inverse_transpose(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Inverse transpose of covariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)

    assert f.is_vector

    jacobian = MatrixFunction([[i.diff(x[j]) for j in range(tdim)] for i in map])
    return jacobian @ f


def contravariant_inverse_transpose(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Inverse transpose of contravariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    tdim = len(inverse_map)
    f = parse_function_input(f_in)
    if substitute:
        f = f.subs(x, inverse_map)

    assert f.is_vector

    j_inv_t = MatrixFunction([[i.diff(x[j]) for i in inverse_map] for j in range(tdim)])
    j_inv_t /= j_inv_t.det()
    return j_inv_t @ f


def identity_inverse(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Inverse of identity().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return identity(f_in, inverse_map, map, substitute)


def l2_inverse(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Inverse of l2().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return l2(f_in, inverse_map, map, substitute)


def covariant_inverse(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Inverse of covariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return covariant(f_in, inverse_map, map, substitute)


def contravariant_inverse(
    f_in: FunctionInput, map: PointType, inverse_map: PointType, substitute: bool = True
) -> Function:
    """Inverse of contravariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return contravariant(f_in, inverse_map, map, substitute)


def double_covariant_inverse(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> Function:
    """Inverse of double_covariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return double_covariant(f_in, inverse_map, map, substitute)


def double_contravariant_inverse(
    f_in: FunctionInput, map: PointType, inverse_map: PointType, substitute: bool = True
) -> Function:
    """Inverse of double_contravariant().

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return double_contravariant(f_in, inverse_map, map, substitute)


def co_contravariant_inverse(
    f_in: FunctionInput,
    map: PointType,
    inverse_map: PointType,
    substitute: bool = True,
) -> MatrixFunction:
    """Map matrix functions.

    Args:
        f_in: The function
        map: The map from the reference cell to the physical cell
        inverse_map: The map to the reference cell from the physical cell
        substitute: Should the inverse map be substituted in?

    Returns:
        The mapped function
    """
    return co_contravariant(f_in, inverse_map, map, substitute)


def get_mapping(
    mapname: str, inverse: bool = False, transpose: bool = False
) -> Callable[[FunctionInput, PointType, PointType, bool], Function]:
    """Get a mapping.

    Args:
        mapname: The name of the mapping
        inverse: Should the map be inverted
        transpose: Should the map be transposed

    Returns:
        A function that performs the mapping
    """
    function_name = mapname
    if inverse:
        function_name += "_inverse"
    if transpose:
        function_name += "_transpose"

    if function_name not in globals():
        raise ValueError(f"Invalid map name: {function_name}")

    return globals()[function_name]
